@@ -32,6 +32,42 @@
 #include <emmintrin.h>
 #include <liboil/liboilcolorspace.h>
 
+/* Work around non-aligned stack frames (which causes the intristics to crash
+ * by making sure the stack frame is always aligned
+ */
+#if defined(__i386__)
+#define OIL_DEFINE_IMPL_FULL_WRAPPER(sse_name, name, flags, ret, ...)	\
+    ret sse_name(__VA_ARGS__) __attribute__((used));                    \
+    ret sse_name ## _wrap (__VA_ARGS__) {                               \
+        OIL_SSE_WRAPPER_CALL(sse_name);                                 \
+    }									\
+    OIL_DEFINE_IMPL_FULL(sse_name ## _wrap, name, flags);
+
+#define OIL_SSE_WRAPPER_CALL(name)                                      \
+    asm volatile(                                                       \
+            "\n\t"                                                      \
+            "subl $0x10,%%esp\n\t"                                      \
+            "andl $0xfffffff0,%%esp\n\t"                                \
+                                                                        \
+            "movdqu 8(%%ebp),%%xmm0\n\t"                                \
+            "movdqa %%xmm0,(%%esp)\n\t"                                 \
+                                                                        \
+            "call " #name "\n\t"                                        \
+            "movl %%ebp,%%esp\n\t"                                      \
+            : :                                                         \
+            : "eax","ecx","edx","xmm0")
+
+#elif defined(__amd64__)
+
+/* Needed because we call *_wrap. Should get optimized away anyway */ 
+
+#define OIL_DEFINE_IMPL_FULL_WRAPPER(sse_name, name, flags, ret, ...)	\
+    OIL_DEFINE_IMPL_FULL(sse_name, name, flags);
+
+#else
+#error Can't use sse on !i386 and !amd64
+#endif
+
 /* non-SSE2 compositing support */
 #define COMPOSITE_OVER(d,s,m) ((d) + (s) - oil_muldiv_255((d),(m)))
 #define COMPOSITE_ADD(d,s) oil_clamp_255((d) + (s))
@@ -41,20 +77,12 @@
  * the channel value in the low byte.  This means 2 pixels per pass.
  */
 
-union m128_int {
-  __m128i m128;
-  uint64_t ull[2];
-};
+static const __m128i c_sse_8x00ff = 
+        {0x00ff00ff00ff00ffULL, 0x00ff00ff00ff00ffULL};
+static const __m128i c_sse_8x0080 = 
+        {0x0080008000800080ULL, 0x0080008000800080ULL};
 
-static const struct _SSEData {
-  union m128_int sse_8x00ff;
-  union m128_int sse_8x0080;
-} c = {
-    .sse_8x00ff.ull =	{0x00ff00ff00ff00ffULL, 0x00ff00ff00ff00ffULL},
-    .sse_8x0080.ull =	{0x0080008000800080ULL, 0x0080008000800080ULL},
-};
-
-#define MC(x) (c.sse_##x.m128)
+#define MC(x) (c_sse_##x)
 
 /* Shuffles the given value such that the alpha for each pixel appears in each
  * channel of the pixel.
@@ -188,8 +216,12 @@ composite_in_argb_const_src_sse_2pix (uint32_t *dest, const uint32_t *src,
 	COMPOSITE_IN(oil_argb_B(*src), m));
   }
 }
-OIL_DEFINE_IMPL_FULL (composite_in_argb_const_src_sse_2pix,
-    composite_in_argb_const_src, OIL_IMPL_FLAG_SSE2);
+
+OIL_DEFINE_IMPL_FULL_WRAPPER (composite_in_argb_const_src_sse_2pix,
+        composite_in_argb_const_src, OIL_IMPL_FLAG_SSE2,
+	static void, 
+	uint32_t *dest, const uint32_t *src, const uint8_t *mask, int n);
+
 
 static void
 composite_in_argb_const_mask_sse_2pix (uint32_t *dest, const uint32_t *src,
@@ -216,8 +248,10 @@ composite_in_argb_const_mask_sse_2pix (uint32_t *dest, const uint32_t *src,
 	COMPOSITE_IN(oil_argb_B(s), mask[0]));
   }
 }
-OIL_DEFINE_IMPL_FULL (composite_in_argb_const_mask_sse_2pix,
-    composite_in_argb_const_mask, OIL_IMPL_FLAG_SSE2);
+OIL_DEFINE_IMPL_FULL_WRAPPER (composite_in_argb_const_mask_sse_2pix,
+    composite_in_argb_const_mask, OIL_IMPL_FLAG_SSE2,
+    static void, 
+    uint32_t *dest, const uint32_t *src, const uint8_t *mask, int n);
 
 static void
 composite_over_argb_sse_2pix (uint32_t *dest, const uint32_t *src, int n)
@@ -272,8 +306,11 @@ composite_over_argb_const_src_sse_2pix (uint32_t *dest, const uint32_t *src,
     *dest++ = d;
   }
 }
-OIL_DEFINE_IMPL_FULL (composite_over_argb_const_src_sse_2pix,
-    composite_over_argb_const_src, OIL_IMPL_FLAG_SSE2);
+
+OIL_DEFINE_IMPL_FULL_WRAPPER (composite_over_argb_const_src_sse_2pix,
+    composite_over_argb_const_src, OIL_IMPL_FLAG_SSE2,
+    static void, 
+    uint32_t *dest, const uint32_t *src, int n);
 
 static void
 composite_in_over_argb_sse_2pix (uint32_t *dest, const uint32_t *src,
@@ -309,8 +346,11 @@ composite_in_over_argb_sse_2pix (uint32_t *dest, const uint32_t *src,
     *dest++ = d;
   }
 }
-OIL_DEFINE_IMPL_FULL (composite_in_over_argb_sse_2pix, composite_in_over_argb,
-    OIL_IMPL_FLAG_SSE2);
+
+OIL_DEFINE_IMPL_FULL_WRAPPER (composite_in_over_argb_sse_2pix,
+    composite_in_over_argb, OIL_IMPL_FLAG_SSE2,
+    static void, 
+    uint32_t *dest, const uint32_t *src, const uint8_t *mask, int n);
 
 static void
 composite_in_over_argb_const_src_sse_2pix (uint32_t *dest, const uint32_t *src,
@@ -348,8 +388,11 @@ composite_in_over_argb_const_src_sse_2pix (uint32_t *dest, const uint32_t *src,
     *dest++ = d;
   }
 }
-OIL_DEFINE_IMPL_FULL (composite_in_over_argb_const_src_sse_2pix,
-    composite_in_over_argb_const_src, OIL_IMPL_FLAG_SSE2);
+
+OIL_DEFINE_IMPL_FULL_WRAPPER (composite_in_over_argb_const_src_sse_2pix,
+    composite_in_over_argb_const_src, OIL_IMPL_FLAG_SSE2,
+    static void, 
+    uint32_t *dest, const uint32_t *src, const uint8_t *mask, int n);
 
 static void
 composite_in_over_argb_const_mask_sse_2pix (uint32_t *dest, const uint32_t *src,
@@ -387,8 +430,11 @@ composite_in_over_argb_const_mask_sse_2pix (uint32_t *dest, const uint32_t *src,
     *dest++ = d;
   }
 }
-OIL_DEFINE_IMPL_FULL (composite_in_over_argb_const_mask_sse_2pix,
-    composite_in_over_argb_const_mask, OIL_IMPL_FLAG_SSE2);
+
+OIL_DEFINE_IMPL_FULL_WRAPPER (composite_in_over_argb_const_mask_sse_2pix,
+    composite_in_over_argb_const_mask, OIL_IMPL_FLAG_SSE2,
+    static void, 
+    uint32_t *dest, const uint32_t *src, const uint8_t *mask, int n);
 
 static void
 composite_over_u8_sse_2pix (uint8_t *dest, const uint8_t *src, int n)
